import justpy as jp
from typing import Awaitable, Callable, Optional, Union
from .float_element import FloatElement

class Number(FloatElement):

    def __init__(self,
                 label: str = None,
                 *,
                 placeholder: str = None,
                 value: float = None,
                 format: str = None,
                 on_change: Optional[Union[Callable, Awaitable]] = None,
                 ):
        """Number Input Element

        :param label: displayed name for the number input
        :param placeholder: text to show if no value is entered
        :param value: the inital value of the field
        :param format: a string like '%.2f' to format the displayed value
        :param on_change: callback to execute when the input is confirmed by leaving the focus
        """
        view = jp.QInput(
            type='number',
            label=label,
            placeholder=placeholder,
            change=self.handle_change,
        )

        super().__init__(view, value=value, format=format, on_change=on_change)

    def handle_change(self, msg):
        msg['value'] = float(msg['value'])

        return super().handle_change(msg)



import justpy as jp

from .element import Element

class Svg(Element):

    def __init__(self,
                 content: str = '',
                 ):
        """Svg Element

        Displays an svg.

        :param content: the svg definition
        """
        view = jp.Div(style="padding:0;width:100%;height:100%")
        super().__init__(view)
        self.content = content

    @property
    def content(self):
        return self.view.inner_html()

    @content.setter
    def content(self, content: any):
        self.view.components = []
        jp.parse_html(content, a=self.view)

    def set_content(self, content: str):
        self.content = content



from typing import Awaitable, Callable, Optional, Union
import justpy as jp

from ..events import ClickEventArguments, handle_event
from .element import Element


class MenuItem(Element):

    def __init__(self,
                 text: str = '',
                 on_click: Optional[Union[Callable, Awaitable]] = None,
                 ):
        """Menu Item Element

        A menu item to be added to a menu.

        :param text: label of the menu item
        :param on_click: callback to be executed when selecting the menu item
        """
        view = jp.QItem(text=text, clickable=True)

        view.on('click', lambda *_: handle_event(on_click, ClickEventArguments(sender=self), update=self.parent_view))

        super().__init__(view)



from __future__ import annotations
import asyncio
import logging
from typing import List, TYPE_CHECKING
if TYPE_CHECKING:
    from starlette.applications import Starlette
    import justpy as jp
    from .config import Config
    from .elements.page import Page

app: 'Starlette'
config: 'Config'
page_stack: List['Page'] = []
view_stack: List['jp.HTMLBaseComponent'] = []
tasks: List[asyncio.tasks.Task] = []
log: logging.Logger = logging.getLogger('nicegui')



#!/usr/bin/env python3
import asyncio
from collections import defaultdict
from justpy.htmlcomponents import HTMLBaseComponent

bindings = defaultdict(list)
bindable_properties = dict()
active_links = []

async def loop():
    while True:
        visited = set()
        invalidated_views = []
        for link in active_links:
            (source_obj, source_name, target_obj, target_name, transform) = link
            value = transform(getattr(source_obj, source_name))
            if getattr(target_obj, target_name) != value:
                setattr(target_obj, target_name, value)
                propagate(target_obj, target_name, visited)
                if hasattr(target_obj, 'view') and isinstance(target_obj.view, HTMLBaseComponent):
                    invalidated_views.append(target_obj.view)
        for view in invalidated_views:
            await view.update()
        await asyncio.sleep(0.1)

def propagate(source_obj, source_name, visited=None):
    if visited is None:
        visited = set()
    visited.add((id(source_obj), source_name))
    for _, target_obj, target_name, transform in bindings[(id(source_obj), source_name)]:
        if (id(target_obj), target_name) in visited:
            continue
        target_value = transform(getattr(source_obj, source_name))
        if getattr(target_obj, target_name) != target_value:
            setattr(target_obj, target_name, target_value)
            propagate(target_obj, target_name, visited)

def bind_to(self_obj, self_name, other_obj, other_name, forward):
    bindings[(id(self_obj), self_name)].append((self_obj, other_obj, other_name, forward))
    if (id(self_obj), self_name) not in bindable_properties:
        active_links.append((self_obj, self_name, other_obj, other_name, forward))
    propagate(self_obj, self_name)

def bind_from(self_obj, self_name, other_obj, other_name, backward):
    bindings[(id(other_obj), other_name)].append((other_obj, self_obj, self_name, backward))
    if (id(other_obj), other_name) not in bindable_properties:
        active_links.append((other_obj, other_name, self_obj, self_name, backward))
    propagate(other_obj, other_name)

class BindableProperty:

    def __set_name__(self, _, name):
        self.name = name

    def __get__(self, owner, _=None):
        return getattr(owner, '_' + self.name)

    def __set__(self, owner, value):
        setattr(owner, '_' + self.name, value)
        bindable_properties[(id(owner), self.name)] = owner
        propagate(owner, self.name)



import justpy as jp
from .group import Group

class Card(Group):

    def __init__(self):
        """Card Element

        Provides a container with a dropped shadow.
        """
        view = jp.QCard(classes='column items-start q-pa-md', style='gap: 1em', delete_flag=False)
        super().__init__(view)

class CardSection(Group):

    def __init__(self):
        view = jp.QCardSection(delete_flag=False)
        super().__init__(view)



import justpy as jp
from typing import Awaitable, Callable, Optional, Union
from .value_element import ValueElement

class StringElement(ValueElement):

    def __init__(self,
                 view: jp.HTMLBaseComponent,
                 *,
                 value: float,
                 on_change: Optional[Union[Callable, Awaitable]],
                 ):
        super().__init__(view, value=value, on_change=on_change)



import traceback
from typing import Awaitable, Callable, Dict, Optional, Union

from ..events import KeyEventArguments, KeyboardAction, KeyboardKey, KeyboardModifiers, handle_event
from .custom_view import CustomView
from .element import Element

class KeyboardView(CustomView):

    def __init__(self, on_key: Callable, repeating: bool):
        super().__init__('keyboard', __file__, active_js_events=['keydown', 'keyup', 'keypress'], repeating=repeating)
        self.allowed_events = ['keyboardEvent']
        self.style = 'display: none'
        self.initialize(temp=False, on_keyboardEvent=on_key)


class Keyboard(Element):

    def __init__(self,
                 *,
                 on_key: Optional[Union[Callable, Awaitable]] = None,
                 active: bool = True,
                 repeating: bool = True,
                 ):
        """
        Keyboard

        Adds global keyboard event tracking.

        :param handle_keys: callback to be executed when keyboard events occur.
        :param active: boolean flag indicating whether the callback should be executed or not (default: True)
        :param repeating: boolean flag indicating whether held keys should be sent repeatedly (default: True)
        """
        super().__init__(KeyboardView(on_key=self.handle_key, repeating=repeating))
        self.active = active
        self.key_handler = on_key

    def handle_key(self, msg: Dict):
        if not self.active:
            return

        try:
            action = KeyboardAction(
                keypress=msg.key_data.action == 'keypress',
                keydown=msg.key_data.action == 'keydown',
                keyup=msg.key_data.action == 'keyup',
                repeat=msg.key_data.repeat,
            )
            modifiers = KeyboardModifiers(
                alt=msg.key_data.altKey,
                ctrl=msg.key_data.ctrlKey,
                meta=msg.key_data.metaKey,
                shift=msg.key_data.shiftKey,
            )
            key = KeyboardKey(
                name=msg.key_data.key,
                code=msg.key_data.code,
                location=msg.key_data.location,
            )
            arguments = KeyEventArguments(sender=self, action=action, modifiers=modifiers, key=key)
            handle_event(self.key_handler, arguments, update=self.parent_view)
        except Exception:
            traceback.print_exc()



from __future__ import annotations
from typing import List
import justpy as jp
from ..globals import view_stack
from ..binding import active_links, bindings, bindable_properties
from .element import Element

class Group(Element):

    def __enter__(self):
        view_stack.append(self.view)
        return self

    def __exit__(self, *_):
        view_stack.pop()

    def tight(self) -> Group:
        return self.classes(replace='').style(replace='')

    def clear(self):
        def collect_components(view: jp.HTMLBaseComponent) -> List[jp.HTMLBaseComponent]:
            return view.components + [view for child in view.components for view in collect_components(child)]
        components = collect_components(self.view)

        active_links[:] = [
            (source_obj, source_name, target_obj, target_name, transform)
            for source_obj, source_name, target_obj, target_name, transform in active_links
            if not (
                isinstance(source_obj, jp.HTMLBaseComponent) and source_obj in components or
                isinstance(target_obj, jp.HTMLBaseComponent) and target_obj in components or
                isinstance(source_obj, Element) and source_obj.view in components or
                isinstance(target_obj, Element) and target_obj.view in components
            )
        ]

        for key, binding_list in list(bindings.items()):
            binding_list[:] = [
                (source_obj, target_obj, target_name, transform)
                for source_obj, target_obj, target_name, transform in binding_list
                if not (
                    isinstance(source_obj, jp.HTMLBaseComponent) and source_obj in components or
                    isinstance(target_obj, jp.HTMLBaseComponent) and target_obj in components or
                    isinstance(source_obj, Element) and source_obj.view in components or
                    isinstance(target_obj, Element) and target_obj.view in components
                )
            ]
            if not binding_list:
                del bindings[key]

        for (obj_id, name), obj in list(bindable_properties.items()):
            if isinstance(obj, jp.HTMLBaseComponent) and obj in components or \
               isinstance(obj, Element) and obj.view in components:
                del bindable_properties[(obj_id, name)]

        self.view.delete_components()



import justpy as jp
from .group import Group

class Menu(Group):

    def __init__(self,
                 *,
                 value: bool = False,
                 ):
        """Menu

        Creates a menu.

        :param value: whether the menu is already opened (default: False)
        """
        view = jp.QMenu(value=value)

        super().__init__(view)

    def open(self):
        self.view.value = True

    def close(self):
        self.view.value = False



import justpy as jp

from .element import Element


class MenuSeparator(Element):

    def __init__(self):
        """Menu Item Separator

        A separator for menus.
        """
        view = jp.QSeparator()
        super().__init__(view)



import justpy as jp
from typing import Optional
from pygments.formatters import HtmlFormatter
from ..globals import config, page_stack, view_stack

class Page(jp.QuasarPage):

    def __init__(self, route: str, title: Optional[str] = None, favicon: Optional[str] = None,
                 classes: str = 'q-ma-md column items-start',
                 css: str = HtmlFormatter().get_style_defs('.codehilite')):
        """Page

        Creates a new page at the given path.

        :param route: the route of the new page. All paths must start with '/', otherwise an error occurs.
        """
        super().__init__()

        self.delete_flag = False
        self.title = title or config.title
        self.favicon = favicon or config.favicon

        self.tailwind = True  # use Tailwind classes instead of Quasars
        self.css = css
        self.head_html += '''
            <script>
                confirm = () => { setTimeout(location.reload.bind(location), 100); return false; };
            </script>
        '''  # avoid confirmation dialog for reload

        self.view = jp.Div(a=self, classes=classes, style='row-gap: 1em')
        self.view.add_page(self)

        jp.Route(route, lambda: self)

    def __enter__(self):
        page_stack.append(self)
        view_stack.append(self.view)
        return self

    def __exit__(self, *_):
        page_stack.pop()
        view_stack.pop()



from typing import Awaitable, Callable, List, Union

startup_tasks: List[Union[Callable, Awaitable]] = []

def on_startup(self, task: Union[Callable, Awaitable]):
    self.startup_tasks.append(task)

shutdown_tasks: List[Union[Callable, Awaitable]] = []

def on_shutdown(self, task: Union[Callable, Awaitable]):
    self.shutdown_tasks.append(task)



import justpy as jp
from .element import Element
from ..task_logger import create_task


class Notify(Element):

    def __init__(self,
                 message: str,
                 *,
                 position: str = 'bottom',
                 close_button: str = None,
                 ):
        """Notification

        Displays a notification on the screen.

        :param message: content of the notification
        :param position: position on the screen ("top-left", "top-right", "bottom-left","bottom-right, "top", "bottom", "left", "right" or "center", default: "bottom")
        :param close_button: optional label of a button to dismiss the notification (default: None)
        """
        view = jp.QNotify(message=message, position=position, closeBtn=close_button)

        super().__init__(view)
        create_task(self.notify_async(), name='notify_async')

    async def notify_async(self):
        self.view.notify = True
        await self.parent_view.update()
        self.view.notify = False



from __future__ import annotations
from typing import List, Optional
from .scene_object3d import Object3D

class Scene(Object3D):

    def __init__(self, view):
        super().__init__('scene', view)

class Group(Object3D):

    def __init__(self):
        super().__init__('group')

class Box(Object3D):

    def __init__(self,
                 width: float = 1.0,
                 height: float = 1.0,
                 depth: float = 1.0,
                 wireframe: bool = False,
                 ):
        super().__init__('box', width, height, depth, wireframe)

class Sphere(Object3D):

    def __init__(self,
                 radius: float = 1.0,
                 width_segments: int = 32,
                 height_segments: int = 16,
                 wireframe: bool = False,
                 ):
        super().__init__('sphere', radius, width_segments, height_segments, wireframe)

class Cylinder(Object3D):

    def __init__(self,
                 top_radius: float = 1.0,
                 bottom_radius: float = 1.0,
                 height: float = 1.0,
                 radial_segments: int = 8,
                 height_segments: int = 1,
                 wireframe: bool = False,
                 ):
        super().__init__('cylinder', top_radius, bottom_radius, height, radial_segments, height_segments, wireframe)

class Extrusion(Object3D):

    def __init__(self,
                 outline: List[List[float, float]],
                 height: float,
                 wireframe: bool = False,
                 ):
        super().__init__('extrusion', outline, height, wireframe)

class Stl(Object3D):

    def __init__(self,
                 url: str,
                 wireframe: bool = False,
                 ):
        super().__init__('stl', url, wireframe)

class Line(Object3D):

    def __init__(self,
                 start: List[float, float, float],
                 end: List[float, float, float],
                 ):
        super().__init__('line', start, end)

class Curve(Object3D):

    def __init__(self,
                 start: List[float, float, float],
                 control1: List[float, float, float],
                 control2: List[float, float, float],
                 end: List[float, float, float],
                 num_points: int = 20,
                 ):
        super().__init__('curve', start, control1, control2, end, num_points)

class Texture(Object3D):

    def __init__(self,
                 url: str,
                 coordinates: List[List[Optional[List[float]]]],
                 ):
        super().__init__('texture', url, coordinates)



from starlette.routing import Route
from . import globals

def add_route(self, route):
    """
    :param route: starlette route including a path and a function to be called
    :return:
    """
    globals.app.routes.insert(0, route)

def get(self, path: str):
    """
    Use as a decorator for a function like @ui.get('/another/route/{id}').
    :param path: string that starts with a '/'
    :return:
    """
    def decorator(func):
        self.add_route(Route(path, func))
        return func
    return decorator



from typing import Awaitable, Callable, Optional, Union
import justpy as jp

from ..binding import bind_from, bind_to, BindableProperty
from ..events import ClickEventArguments, handle_event
from .element import Element

class Button(Element):
    text = BindableProperty()

    def __init__(self,
                 text: str = '',
                 *,
                 on_click: Optional[Union[Callable, Awaitable]] = None,
                 ):
        """Button Element

        :param text: the label of the button
        :param on_click: callback which is invoked when button is pressed
        """

        view = jp.QButton(label=text, color='primary')
        super().__init__(view)

        self.text = text
        self.bind_text_to(self.view, 'label')

        def process_event(view, event):
            socket = event.get('websocket')
            handle_event(on_click, ClickEventArguments(sender=self, socket=socket), update=self.parent_view)

        view.on('click', process_event)

    def set_text(self, text: str):
        self.text = text

    def bind_text_to(self, target_object, target_name, forward=lambda x: x):
        bind_to(self, 'text', target_object, target_name, forward=forward)
        return self

    def bind_text_from(self, target_object, target_name, backward=lambda x: x):
        bind_from(self, 'text', target_object, target_name, backward=backward)
        return self

    def bind_text(self, target_object, target_name, forward=lambda x: x, backward=lambda x: x):
        bind_from(self, 'text', target_object, target_name, backward=backward)
        bind_to(self, 'text', target_object, target_name, forward=forward)
        return self



import justpy as jp
from ..binding import bind_from, bind_to, BindableProperty
from ..globals import view_stack, page_stack

class Element:
    visible = BindableProperty()

    def __init__(self,
                 view: jp.HTMLBaseComponent,
                 ):
        self.parent_view = view_stack[-1]
        self.parent_view.add(view)
        self.view = view
        self.page = page_stack[-1]
        self.view.add_page(self.page)

        self.visible = True

    @property
    def visible(self):
        return self.visible_

    @visible.setter
    def visible(self, visible: bool):
        self.visible_ = visible
        (self.view.remove_class if self.visible_ else self.view.set_class)('hidden')

    def bind_visibility_to(self, target_object, target_name, forward=lambda x: x):
        bind_to(self, 'visible', target_object, target_name, forward=forward)
        return self

    def bind_visibility_from(self, target_object, target_name, backward=lambda x: x, *, value=None):
        if value is not None:
            def backward(x): return x == value

        bind_from(self, 'visible', target_object, target_name, backward=backward)
        return self

    def bind_visibility(self, target_object, target_name, forward=lambda x: x, backward=None, *, value=None):
        if value is not None:
            def backward(x): return x == value

        bind_from(self, 'visible', target_object, target_name, backward=backward)
        bind_to(self, 'visible', target_object, target_name, forward=forward)
        return self

    def classes(self, add: str = None, *, remove: str = None, replace: str = None):
        '''HTML classes to modify the look of the element.
        Every class in the `remove` parameter will be removed from the element.
        Classes are seperated with a blank space.
        This can be helpful if the predefined classes by NiceGUI are not wanted in a particular styling.
        '''
        class_list = [] if replace is not None else self.view.classes.split()
        class_list = [c for c in class_list if c not in (remove or '')]
        class_list += (add or '').split()
        class_list += (replace or '').split()
        self.view.classes = ' '.join(class_list)

        return self

    def style(self, add: str = None, *, remove: str = None, replace: str = None):
        '''CSS style sheet definitions to modify the look of the element.
        Every style in the `remove` parameter will be removed from the element.
        Styles are seperated with a semicolon.
        This can be helpful if the predefined style sheet definitions by NiceGUI are not wanted in a particular styling.
        '''
        style_list = [] if replace is not None else self.view.style.split(';')
        style_list = [c for c in style_list if c not in (remove or '').split(';')]
        style_list += (add or '').split(';')
        style_list += (replace or '').split(';')
        self.view.style = ';'.join(style_list)

        return self

    def props(self, add: str = None, *, remove: str = None, replace: str = None):
        '''Quasar props https://quasar.dev/vue-components/button#design to modify the look of the element.
        Boolean props will automatically activated if they appear in the list of the `add` property.
        Props are seperated with a blank space.
        Every prop passed to the `remove` parameter will be removed from the element.
        This can be helpful if the predefined props by NiceGUI are not wanted in a particular styling.
        '''
        for prop in (remove or '').split() + (replace or '').split():
            setattr(self.view, prop.split('=')[0], None)

        for prop in (add or '').split() + (replace or '').split():
            if '=' in prop:
                setattr(self.view, *prop.split('='))
            else:
                setattr(self.view, prop, True)

        return self



import asyncio
from inspect import signature
from justpy.htmlcomponents import HTMLBaseComponent
from pydantic import BaseModel
import traceback
from typing import Any, Awaitable, Callable, List, Optional, Union
from starlette.websockets import WebSocket

from .elements.element import Element
from .task_logger import create_task

class EventArguments(BaseModel):
    class Config:
        arbitrary_types_allowed = True
    sender: Element
    socket: Optional[WebSocket]

class ClickEventArguments(EventArguments):
    pass

class UploadEventArguments(EventArguments):
    files: List[bytes]

class ValueChangeEventArguments(EventArguments):
    value: Any

class KeyboardAction(BaseModel):
    keypress: bool
    keydown: bool
    keyup: bool
    repeat: bool

class KeyboardModifiers(BaseModel):
    alt: bool
    ctrl: bool
    meta: bool
    shift: bool

class KeyboardKey(BaseModel):
    name: str
    code: str
    location: int

    def __eq__(self, other: str) -> bool:
        return self.name == other or self.code == other

    def __repr__(self):
        return str(self.name)

    @property
    def is_cursorkey(self):
        return self.code.startswith('Arrow')

    @property
    def number(self) -> Optional[int]:
        """Integer value of a number key."""
        return int(self.code.removeprefix('Digit')) if self.code.startswith('Digit') else None

    @property
    def backspace(self) -> bool:
        return self.name == 'Backspace'

    @property
    def tab(self) -> bool:
        return self.name == 'Tab'

    @property
    def enter(self) -> bool:
        return self.name == 'enter'

    @property
    def shift(self) -> bool:
        return self.name == 'Shift'

    @property
    def control(self) -> bool:
        return self.name == 'Control'

    @property
    def alt(self) -> bool:
        return self.name == 'Alt'

    @property
    def pause(self) -> bool:
        return self.name == 'Pause'

    @property
    def caps_lock(self) -> bool:
        return self.name == 'CapsLock'

    @property
    def escape(self) -> bool:
        return self.name == 'Escape'

    @property
    def space(self) -> bool:
        return self.name == 'Space'

    @property
    def page_up(self) -> bool:
        return self.name == 'PageUp'

    @property
    def page_down(self) -> bool:
        return self.name == 'PageDown'

    @property
    def end(self) -> bool:
        return self.name == 'End'

    @property
    def home(self) -> bool:
        return self.name == 'Home'

    @property
    def arrow_left(self) -> bool:
        return self.name == 'ArrowLeft'

    @property
    def arrow_up(self) -> bool:
        return self.name == 'ArrowUp'

    @property
    def arrow_right(self) -> bool:
        return self.name == 'ArrowRight'

    @property
    def arrow_down(self) -> bool:
        return self.name == 'ArrowDown'

    @property
    def print_screen(self) -> bool:
        return self.name == 'PrintScreen'

    @property
    def insert(self) -> bool:
        return self.name == 'Insert'

    @property
    def delete(self) -> bool:
        return self.name == 'Delete'

    @property
    def meta(self) -> bool:
        return self.name == 'Meta'

    @property
    def f1(self) -> bool:
        return self.name == 'F1'

    @property
    def f2(self) -> bool:
        return self.name == 'F2'

    @property
    def f3(self) -> bool:
        return self.name == 'F3'

    @property
    def f4(self) -> bool:
        return self.name == 'F4'

    @property
    def f5(self) -> bool:
        return self.name == 'F5'

    @property
    def f6(self) -> bool:
        return self.name == 'F6'

    @property
    def f7(self) -> bool:
        return self.name == 'F7'

    @property
    def f8(self) -> bool:
        return self.name == 'F8'

    @property
    def f9(self) -> bool:
        return self.name == 'F9'

    @property
    def f10(self) -> bool:
        return self.name == 'F10'

    @property
    def f11(self) -> bool:
        return self.name == 'F11'

    @property
    def f12(self) -> bool:
        return self.name == 'F12'

class KeyEventArguments(EventArguments):
    class Config:
        arbitrary_types_allowed = True
    action: KeyboardAction
    key: KeyboardKey
    modifiers: KeyboardModifiers


def handle_event(handler: Optional[Union[Callable, Awaitable]],
                 arguments: EventArguments,
                 *,
                 update: Optional[HTMLBaseComponent] = None):
    try:
        if handler is None:
            return
        no_arguments = not signature(handler).parameters
        result = handler() if no_arguments else handler(arguments)
        if asyncio.iscoroutinefunction(handler):
            async def async_handler():
                try:
                    await result
                    if update is not None:
                        await update.update()
                except Exception:
                    traceback.print_exc()
            create_task(async_handler(), name=str(handler))
            return False
        else:
            return result
    except Exception:
        traceback.print_exc()



from .custom_view import CustomView
from .element import Element

class CustomExampleView(CustomView):

    def __init__(self, on_change):
        super().__init__('custom_example', __file__, value=0)

        self.on_change = on_change
        self.allowed_events = ['onAdd']
        self.initialize(temp=False, onAdd=self.handle_add)

    def handle_add(self, msg):
        self.options.value += msg.number
        if self.on_change is not None:
            return self.on_change(self.options.value)
        return False

class CustomExample(Element):

    def __init__(self, *, on_change=None):
        super().__init__(CustomExampleView(on_change))

    def add(self, number: str):
        self.view.options.value += number
        self.view.on_change(self.view.options.value)



import justpy as jp

from ..binding import bind_from, bind_to, BindableProperty
from .group import Group

class Image(Group):
    source = BindableProperty()

    def __init__(self,
                 source: str = '',
                 ):
        """Image Element

        Displays an image.

        :param source: the source of the image; can be an url or a base64 string
        """
        view = jp.QImg(src=source)
        super().__init__(view)

        self.source = source
        self.bind_source_to(self.view, 'src')

    def set_source(self, source: str):
        self.source = source

    def bind_source_to(self, target_object, target_name, forward=lambda x: x):
        bind_to(self, 'source', target_object, target_name, forward=forward)
        return self

    def bind_source_from(self, target_object, target_name, backward=lambda x: x):
        bind_from(self, 'source', target_object, target_name, backward=backward)
        return self

    def bind_source(self, target_object, target_name, forward=lambda x: x, backward=lambda x: x):
        bind_from(self, 'source', target_object, target_name, backward=backward)
        bind_to(self, 'source', target_object, target_name, forward=forward)
        return self



from __future__ import annotations
import asyncio
from typing import List, Optional
import uuid
import numpy as np
from justpy.htmlcomponents import WebPage
from ..task_logger import create_task

class Object3D:
    stack: List[Object3D] = []

    def __init__(self, type: str, *args):
        self.type = type
        self.id = str(uuid.uuid4())
        self.name = None
        self.parent = self.stack[-1]
        self.view = self.parent.view
        self.page = self.parent.page
        self.args = args
        self.color = '#ffffff'
        self.opacity = 1.0
        self.x = 0
        self.y = 0
        self.z = 0
        self.R = [[1, 0, 0], [0, 1, 0], [0, 0, 1]]
        self.sx = 1
        self.sy = 1
        self.sz = 1
        self.run_command(self._create_command)
        self.view.objects[self.id] = self

    def with_name(self, name: str) -> Object3D:
        self.name = name
        return self

    def run_command(self, command: str, socket=None):
        sockets = [socket] if socket else WebPage.sockets.get(self.page.page_id, {}).values()
        for socket in sockets:
            create_task(self.view.run_method(command, socket), name=command)

    def send_to(self, socket):
        self.run_command(self._create_command, socket)
        self.run_command(self._material_command, socket)
        self.run_command(self._move_command, socket)
        self.run_command(self._rotate_command, socket)
        self.run_command(self._scale_command, socket)

    def __enter__(self):
        self.stack.append(self)
        return self

    def __exit__(self, *_):
        self.stack.pop()

    @property
    def _create_command(self):
        return f'create("{self.type}", "{self.id}", "{self.parent.id}", {str(self.args)[1:-1]})'

    @property
    def _material_command(self):
        return f'material("{self.id}", "{self.color}", "{self.opacity}")'

    @property
    def _move_command(self):
        return f'move("{self.id}", {self.x}, {self.y}, {self.z})'

    @property
    def _rotate_command(self):
        return f'rotate("{self.id}", {self.R})'

    @property
    def _scale_command(self):
        return f'scale("{self.id}", {self.sx}, {self.sy}, {self.sz})'

    @property
    def _delete_command(self):
        return f'delete("{self.id}")'

    def material(self, color: str = '#ffffff', opacity: float = 1.0):
        if self.color != color or self.opacity != opacity:
            self.color = color
            self.opacity = opacity
            self.run_command(self._material_command)
        return self

    def move(self, x: float = 0.0, y: float = 0.0, z: float = 0.0) -> Object3D:
        if self.x != x or self.y != y or self.z != z:
            self.x = x
            self.y = y
            self.z = z
            self.run_command(self._move_command)
        return self

    def rotate(self, omega: float, phi: float, kappa: float):
        Rx = np.array([[1, 0, 0], [0, np.cos(omega), -np.sin(omega)], [0, np.sin(omega), np.cos(omega)]])
        Ry = np.array([[np.cos(phi), 0, np.sin(phi)], [0, 1, 0], [-np.sin(phi), 0, np.cos(phi)]])
        Rz = np.array([[np.cos(kappa), -np.sin(kappa), 0], [np.sin(kappa), np.cos(kappa), 0], [0, 0, 1]])
        return self.rotate_R((Rz @ Ry @ Rx).tolist())

    def rotate_R(self, R: List[List[float]]):
        if self.R != R:
            self.R = R
            self.run_command(self._rotate_command)
        return self

    def scale(self, sx: float = 1.0, sy: Optional[float] = None, sz: Optional[float] = None):
        if sy is None:
            sy = sx
        if sz is None:
            sz = sx
        if self.sx != sx or self.sy != sy or self.sz != sz:
            self.sx = sx
            self.sy = sy
            self.sz = sz
            self.run_command(self._scale_command)
        return self

    def delete(self):
        del self.view.objects[self.id]
        self.run_command(self._delete_command)



import traceback
import justpy as jp
from typing import Awaitable, Callable, Optional, Union
import base64

from ..events import UploadEventArguments, handle_event
from .element import Element

class Upload(Element):

    def __init__(self,
                 *,
                 multiple: bool = False,
                 on_upload: Optional[Union[Callable, Awaitable]] = None,
                 ):
        """File Upload Element

        :param multiple: allow uploading multiple files at once (default: False)
        :param on_upload: callback to execute when a file is uploaded (list of bytearrays)
        """
        self.upload_handler = on_upload
        view = jp.Form(enctype='multipart/form-data', classes='flex gap-4 items-center',
                       submit=lambda s, m: self.submit(s, m))
        jp.Input(type='file', multiple=multiple, a=view)
        jp.QButton(type='submit', text='Upload', color='primary', a=view)

        super().__init__(view)

    def submit(self, _, msg):
        try:
            for form_data in msg.form_data:
                if form_data.type == 'file':
                    files = [base64.b64decode(f.file_content) for f in form_data.files]
                    arguments = UploadEventArguments(sender=self, files=files)
                    handle_event(self.upload_handler, arguments, update=self.parent_view)
        except Exception:
            traceback.print_exc()



class Ui:
    from .config import config  # NOTE: before run
    from .run import run  # NOTE: before justpy

    from .elements.button import Button as button
    from .elements.checkbox import Checkbox as checkbox
    from .elements.custom_example import CustomExample as custom_example
    from .elements.dialog import Dialog as dialog
    from .elements.icon import Icon as icon
    from .elements.image import Image as image
    from .elements.input import Input as input
    from .elements.joystick import Joystick as joystick
    from .elements.html import Html as html
    from .elements.label import Label as label
    from .elements.link import Link as link
    from .elements.log import Log as log
    from .elements.markdown import Markdown as markdown
    from .elements.menu import Menu as menu
    from .elements.menu_item import MenuItem as menu_item
    from .elements.menu_separator import MenuSeparator as menu_separator
    from .elements.notify import Notify as notify
    from .elements.number import Number as number
    from .elements.open import open, open_async
    from .elements.page import Page as page
    from .elements.radio import Radio as radio
    from .elements.scene import Scene as scene
    from .elements.select import Select as select
    from .elements.slider import Slider as slider
    from .elements.svg import Svg as svg
    from .elements.switch import Switch as switch
    from .elements.toggle import Toggle as toggle
    from .elements.upload import Upload as upload

    from .elements.plot import Plot as plot
    from .elements.line_plot import LinePlot as line_plot

    from .elements.row import Row as row
    from .elements.column import Column as column
    from .elements.card import Card as card
    from .elements.keyboard import Keyboard as keyboard
    from .elements.card import CardSection as card_section

    from .timer import Timer as timer

    from .lifecycle import startup_tasks, on_startup, shutdown_tasks, on_shutdown

    from .routes import add_route, get



import justpy as jp
from .element import Element

class Html(Element):

    def __init__(self,
                 content: str = '',
                 ):
        """HTML Element

        Renders arbitrary HTML onto the page. `Tailwind <https://tailwindcss.com/>`_ can be used for styling.

        :param content: the HTML code to be displayed
        """
        view = jp.QDiv()
        super().__init__(view)
        self.content = content

    @property
    def content(self) -> str:
        return self.view.inner_html

    @content.setter
    def content(self, content: any):
        self.set_content(content)

    def set_content(self, content: str):
        self.view.inner_html = content



import justpy as jp
from .group import Group

class Row(Group):

    def __init__(self):
        '''Row Element

        Provides a container which arranges its child in a row.
        '''
        view = jp.QDiv(classes='row items-start', style='gap: 1em', delete_flag=False)
        super().__init__(view)



import justpy as jp

from ..binding import bind_from, bind_to, BindableProperty
from .element import Element

class Label(Element):
    text = BindableProperty()

    def __init__(self,
                 text: str = '',
                 ):
        """Label Element

        Displays some text.

        :param text: the content of the label
        """
        view = jp.Div(text=text)
        super().__init__(view)

        self.text = text
        self.bind_text_to(self.view, 'text')

    def set_text(self, text: str):
        self.text = text

    def bind_text_to(self, target_object, target_name, forward=lambda x: x):
        bind_to(self, 'text', target_object, target_name, forward=forward)
        return self

    def bind_text_from(self, target_object, target_name, backward=lambda x: x):
        bind_from(self, 'text', target_object, target_name, backward=backward)
        return self

    def bind_text(self, target_object, target_name, forward=lambda x: x, backward=lambda x: x):
        bind_from(self, 'text', target_object, target_name, backward=backward)
        bind_to(self, 'text', target_object, target_name, forward=forward)
        return self



import justpy as jp
from typing import Any, Awaitable, Callable, Dict, List, Optional, Union
from .value_element import ValueElement

class ChoiceElement(ValueElement):

    def __init__(self,
                 view: jp.HTMLBaseComponent,
                 options: Union[List, Dict],
                 *,
                 value: Any,
                 on_change: Optional[Union[Callable, Awaitable]] = None,
                 ):
        if isinstance(options, List):
            view.options = [{'label': option, 'value': option} for option in options]
        else:
            view.options = [{'label': value, 'value': key} for key, value in options.items()]

        super().__init__(view, value=value, on_change=on_change)



import justpy as jp
from typing import Any, Awaitable, Callable, Optional, Union

from ..events import ValueChangeEventArguments, handle_event
from ..binding import bind_from, bind_to, BindableProperty
from .element import Element

class ValueElement(Element):
    value = BindableProperty()

    def __init__(self,
                 view: jp.HTMLBaseComponent,
                 *,
                 value: Any,
                 on_change: Optional[Union[Callable, Awaitable]],
                 ):
        super().__init__(view)

        self.change_handler = on_change
        self.value = value
        self.bind_value_to(self.view, 'value', forward=self.value_to_view)

    def value_to_view(self, value):
        return value

    def handle_change(self, msg):
        self.value = msg['value']
        arguments = ValueChangeEventArguments(sender=self, value=self.value)
        handle_event(self.change_handler, arguments, update=self.parent_view)

    def bind_value_to(self, target_object, target_name, *, forward=lambda x: x):
        bind_to(self, 'value', target_object, target_name, forward=forward)
        return self

    def bind_value_from(self, target_object, target_name, *, backward=lambda x: x):
        bind_from(self, 'value', target_object, target_name, backward=backward)
        return self

    def bind_value(self, target_object, target_name, *, forward=lambda x: x, backward=lambda x: x):
        bind_from(self, 'value', target_object, target_name, backward=backward)
        bind_to(self, 'value', target_object, target_name, forward=forward)
        return self



import justpy as jp
from typing import Awaitable, Callable, Dict, List, Optional, Union
from .choice_element import ChoiceElement

class Radio(ChoiceElement):

    def __init__(self,
                 options: Union[List, Dict],
                 *,
                 value: any = None,
                 on_change: Optional[Union[Callable, Awaitable]] = None,
                 ):
        """Radio Selection Element

        :param options: a list ['value1', ...] or dictionary `{'value1':'label1', ...}` specifying the options
        :param value: the inital value
        :param on_change: callback to execute when selection changes
        """
        view = jp.QOptionGroup(options=options, input=self.handle_change)

        super().__init__(view, options, value=value, on_change=on_change)



#!/usr/bin/env python3
from typing import Awaitable, Callable
import asyncio

from .ui import Ui  # NOTE: before justpy
import justpy as jp
from .timer import Timer
from . import globals
from . import binding
from .task_logger import create_task


@jp.app.on_event('startup')
def startup():
    globals.tasks.extend(create_task(t.coro, name=t.name) for t in Timer.prepared_coroutines)
    Timer.prepared_coroutines.clear()
    globals.tasks.extend(create_task(t, name='startup task') for t in Ui.startup_tasks if isinstance(t, Awaitable))
    [safe_invoke(t) for t in Ui.startup_tasks if isinstance(t, Callable)]
    jp.run_task(binding.loop())

@jp.app.on_event('shutdown')
def shutdown():
    [create_task(t, name='shutdown task') for t in Ui.shutdown_tasks if isinstance(t, Awaitable)]
    [safe_invoke(t) for t in Ui.shutdown_tasks if isinstance(t, Callable)]
    [t.cancel() for t in globals.tasks]

def safe_invoke(func: Callable):
    try:
        func()
    except:
        globals.log.exception(f'could not invoke {func}')

app = globals.app = jp.app
ui = Ui()

page = ui.page('/')
page.__enter__()
jp.justpy(lambda: page, start_server=False)



import inspect
import sys
import webbrowser
import uvicorn
from . import globals

if not globals.config.interactive and globals.config.reload and not inspect.stack()[-2].filename.endswith('spawn.py'):

    if globals.config.show:
        webbrowser.open(f'http://{globals.config.host}:{globals.config.port}/')
    uvicorn.run('nicegui:app', host=globals.config.host, port=globals.config.port, lifespan='on', reload=True,
                log_level=globals.config.uvicorn_logging_level)
    sys.exit()

def run(self, *,
        host: str = '0.0.0.0',
        port: int = 8080,
        title: str = 'NiceGUI',
        favicon: str = 'favicon.ico',
        reload: bool = True,
        show: bool = True,
        uvicorn_logging_level: str = 'warning',
        ):

    if globals.config.interactive or reload == False:  # NOTE: if reload == True we already started uvicorn above
        if show:
            webbrowser.open(f'http://{host if host != "0.0.0.0" else "127.0.0.1"}:{port}/')
        uvicorn.run(globals.app, host=host, port=port, lifespan='on', log_level=uvicorn_logging_level)



from typing import Awaitable, Callable, Optional, Union
import traceback
import websockets
from .element import Element
from .custom_view import CustomView
from .page import Page
from .scene_object3d import Object3D
from ..globals import view_stack
from ..events import handle_event

class SceneView(CustomView):

    def __init__(self, *, width: int, height: int, on_click: Union[Callable, Awaitable]):
        dependencies = ['three.min.js', 'OrbitControls.js', 'STLLoader.js']
        super().__init__('scene', __file__, dependencies, width=width, height=height)
        self.on_click = on_click
        self.allowed_events = ['onConnect', 'onClick']
        self.initialize(temp=False, onConnect=self.handle_connect, onClick=self.handle_click)
        self.objects = {}

    def handle_connect(self, msg):
        try:
            for object in self.objects.values():
                object.send_to(msg.websocket)
        except:
            traceback.print_exc()

    def handle_click(self, msg):
        try:
            for hit in msg.hits:
                hit.object = self.objects.get(hit.object_id)
            handle_event(self.on_click, msg, update=self)
            return False
        except:
            traceback.print_exc()

    async def run_method(self, command, websocket):
        try:
            await websocket.send_json({'type': 'run_method', 'data': command, 'id': self.id})
        except (websockets.exceptions.ConnectionClosedOK, RuntimeError):
            pass
        return True

class Scene(Element):
    from .scene_objects import Group as group
    from .scene_objects import Box as box
    from .scene_objects import Sphere as sphere
    from .scene_objects import Cylinder as cylinder
    from .scene_objects import Extrusion as extrusion
    from .scene_objects import Stl as stl
    from .scene_objects import Line as line
    from .scene_objects import Curve as curve
    from .scene_objects import Texture as texture

    def __init__(self, width: int = 400, height: int = 300, on_click: Optional[Union[Callable, Awaitable]] = None):
        """3D Scene

        Display a 3d scene using `three.js <https://threejs.org/>`_.
        Currently NiceGUI supports boxes, spheres, cylinders/cones, extrusions, straight lines, curves and textured meshes.
        Objects can be translated, rotated and displayed with different color, opacity or as wireframes.
        They can also be grouped to apply joint movements.

        :param width: width of the canvas
        :param height: height of the canvas
        :param on_click: callback to execute when a 3d object is clicked
        """
        super().__init__(SceneView(width=width, height=height, on_click=on_click))

    def __enter__(self):
        view_stack.append(self.view)
        scene = self.view.objects.get('scene', SceneObject(self.view, self.page))
        Object3D.stack.clear()
        Object3D.stack.append(scene)
        return self

    def __exit__(self, *_):
        view_stack.pop()

class SceneObject:

    def __init__(self, view: SceneView, page: Page):
        self.id = 'scene'
        self.view = view
        self.page = page



import justpy as jp
from typing import Awaitable, Callable, Optional, Union
from .string_element import StringElement

class Input(StringElement):

    def __init__(self,
                 label: str = None,
                 *,
                 placeholder: str = None,
                 value: str = '',
                 on_change: Optional[Union[Callable, Awaitable]] = None,
                 ):
        """Text Input Element

        :param label: displayed label for the text input
        :param placeholder: text to show if no value is entered
        :param value: the current value of the text input
        :param on_change: callback to execute when the input is confirmed by leaving the focus
        """
        view = jp.QInput(
            label=label,
            placeholder=placeholder,
            value=value,
            change=self.handle_change,
        )

        super().__init__(view, value=value, on_change=on_change)



from typing import Callable, Dict, Optional
from .custom_view import CustomView
from .element import Element

class JoystickView(CustomView):

    def __init__(self,
                 on_start: Optional[Callable],
                 on_move: Optional[Callable],
                 on_end: Optional[Callable],
                 **options,
                 ):
        super().__init__('joystick', __file__, ['nipplejs.min.js'], **options)

        self.on_start = on_start
        self.on_move = on_move
        self.on_end = on_end
        self.allowed_events = ['onStart', 'onMove', 'onEnd']
        self.initialize(temp=False,
                        onStart=self.handle_start,
                        onMove=self.handle_move,
                        onEnd=self.handle_end)

    def handle_start(self, msg):
        if self.on_start is not None:
            return self.on_start(msg)
        return False

    def handle_move(self, msg):
        if self.on_move is not None:
            return self.on_move(msg)
        return False

    def handle_end(self, msg):
        if self.on_end is not None:
            return self.on_end(msg)
        return False

class Joystick(Element):

    def __init__(self,
                 *,
                 on_start: Optional[Callable] = None,
                 on_move: Optional[Callable] = None,
                 on_end: Optional[Callable] = None,
                 **options: Dict,
                 ):
        """Joystick

        Create a joystick based on `nipple.js <https://yoannmoi.net/nipplejs/>`_.

        :param on_start: callback for when the user toches the joystick
        :param on_move: callback for when the user moves the joystick
        :param on_end: callback for when the user releases the joystick
        :param options: arguments like `color` which should be passed to the `underlying nipple.js library <https://github.com/yoannmoinet/nipplejs#options>`_
        """
        super().__init__(JoystickView(on_start, on_move, on_end, **options))



import asyncio
import time
import traceback
from typing import Awaitable, Callable, List, Union
from collections import namedtuple

from .binding import BindableProperty
from .globals import tasks, view_stack
from .task_logger import create_task

NamedCoroutine = namedtuple('NamedCoroutine', ['name', 'coro'])

class Timer:
    prepared_coroutines: List[NamedCoroutine] = []

    active = BindableProperty()

    def __init__(self, interval: float, callback: Union[Callable, Awaitable], *, active: bool = True, once: bool = False):
        """Timer

        One major drive behind the creation of NiceGUI was the necessity to have a simple approach to update the interface in regular intervals, for example to show a graph with incomming measurements.
        A timer will execute a callback repeatedly with a given interval.
        The parent view container will be updated automatically, as long as the callback does not return `False`.

        :param interval: the interval in which the timer is called
        :param callback: function or coroutine to execute when interval elapses (can return `False` to prevent view update)
        :param active: whether the callback should be executed or not
        :param once: whether the callback is only executed once after a delay specified by `interval`; default is `False`
        """

        parent = view_stack[-1]
        self.active = active

        async def do_callback():
            try:
                if asyncio.iscoroutinefunction(callback):
                    return await callback()
                else:
                    return callback()
            except Exception:
                traceback.print_exc()

        async def timeout():
            await asyncio.sleep(interval)
            await do_callback()
            await parent.update()

        async def loop():
            while True:
                try:
                    start = time.time()
                    if self.active:
                        needs_update = await do_callback()
                        if needs_update != False:
                            await parent.update()
                    dt = time.time() - start
                    await asyncio.sleep(interval - dt)
                except asyncio.CancelledError:
                    pass
                except:
                    traceback.print_exc()
                    await asyncio.sleep(interval)

        coroutine = timeout() if once else loop()
        event_loop = asyncio.get_event_loop()
        if not event_loop.is_running():
            self.prepared_coroutines.append(NamedCoroutine(str(callback), coroutine))
        else:
            tasks.append(create_task(coroutine, name=str(callback)))



import justpy as jp
from .element import Element

class Icon(Element):

    def __init__(self,
                 name: str,
                 ):
        view = jp.QIcon(name=name, classes=f'q-pt-xs', size='20px')

        super().__init__(view)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['nicegui', 'nicegui.elements']

package_data = \
{'': ['*'],
 'nicegui': ['static/*',
             'static/templates/*',
             'static/templates/css/*',
             'static/templates/highcharts/*',
             'static/templates/js/*',
             'static/templates/local/*',
             'static/templates/local/fontawesome/css/*',
             'static/templates/local/fontawesome/webfonts/*',
             'static/templates/local/ionicons/*',
             'static/templates/local/materialdesignicons/iconfont/*',
             'static/templates/local/robotofont/*',
             'static/templates/local/tailwind/*'],
 'nicegui.elements': ['lib/*']}

install_requires = \
['Pygments>=2.9.0,<3.0.0',
 'asttokens>=2.0.5,<3.0.0',
 'docutils>=0.17.1,<0.18.0',
 'justpy==0.1.5',
 'markdown2>=2.4.0,<3.0.0',
 'matplotlib>=3.4.1,<4.0.0',
 'pydantic>=1.8.2,<2.0.0',
 'typing-extensions>=3.10.0,<4.0.0',
 'uvicorn[watchgodreloader]>=0.14.0,<0.15.0',
 'watchgod>=0.7,<0.8',
 'websockets>=9.1,<10.0']

setup_kwargs = {
    'name': 'nicegui',
    'version': '0.7.1',
    'description': 'Web User Interface with Buttons, Dialogs, Markdown, 3D Scences and Plots',
    'long_description': '# NiceGUI\n\n<img src="https://raw.githubusercontent.com/zauberzeug/nicegui/main/sceenshots/ui-elements.png" width="300" align="right">\n\nNiceGUI is an easy-to-use, Python-based UI framework, which renders to the web browser.\nYou can create buttons, dialogs, markdown, 3D scenes, plots and much more.\n\nIt was designed to be used for micro web apps, dashboards, robotics projects, smart home solutions and similar use cases.\nIt is also helpful for development, for example when tweaking/configuring a machine learning algorithm or tuning motor controllers.\n\n## Features\n\n- browser-based graphical user interface\n- shared state between multiple browser windows\n- implicit reload on code change\n- standard GUI elements like label, button, checkbox, switch, slider, input, file upload, ...\n- simple grouping with rows, columns, cards and dialogs\n- general-purpose HTML and markdown elements\n- powerful elements to plot graphs, render 3D scenes and get steering events via virtual joysticks\n- built-in timer to refresh data in intervals (even every 10 ms)\n- straight-forward data binding to write even less code\n- notifications, dialogs and menus to provide state of the art user interaction\n\n## Installation\n\n```bash\npython3 -m pip install nicegui\n```\n\n## Usage\n\nWrite your nice GUI in a file `main.py`:\n\n```python\nfrom nicegui import ui\n\nui.label(\'Hello NiceGUI!\')\nui.button(\'BUTTON\', on_click=lambda: print(\'button was pressed\', flush=True))\n\nui.run()\n```\n\nLaunch it with:\n\n```bash\npython3 main.py\n```\n\nThe GUI is now available through http://localhost:8080/ in your browser.\nNote: The script will automatically reload the page when you modify the code.\n\nFull documentation can be found at [https://nicegui.io](https://nicegui.io).\n\n## Configuration\n\nYou can call `ui.run()` with optional arguments for some high-level configuration:\n\n- `host` (default: `\'0.0.0.0\'`)\n- `port` (default: `8080`)\n- `title` (default: `\'NiceGUI\'`)\n- `favicon` (default: `\'favicon.ico\'`)\n- `reload`: automatically reload the ui on file changes (default: `True`)\n- `show`: automatically open the ui in a browser tab (default: `True`)\n- `uvicorn_logging_level`: logging level for uvicorn server (default: `\'warning\'`)\n- `interactive`: used internally when run in interactive Python shell (default: `False`)\n\n## Docker\n\nUse the [multi-arch docker image](https://hub.docker.com/repository/docker/zauberzeug/nicegui) for pain-free installation:\n\n```bash\ndocker run --rm -p 8888:8080 -v $(pwd)/my_script.py:/app/main.py -it zauberzeug/nicegui:latest\n```\n\nThis will start the server at http://localhost:8888 with code from `my_script.py` within the current directory.\nCode modification triggers an automatic reload.\n\n## Why?\n\nWe like [Streamlit](https://streamlit.io/) but find it does [too much magic when it comes to state handling](https://github.com/zauberzeug/nicegui/issues/1#issuecomment-847413651).\nIn search for an alternative nice library to write simple graphical user interfaces in Python we discovered [justpy](https://justpy.io/).\nWhile too "low-level HTML" for our daily usage it provides a great basis for "NiceGUI".\n\n## API\n\nThe API reference is hosted at [https://nicegui.io](https://nicegui.io) and is [implemented with NiceGUI itself](https://github.com/zauberzeug/nicegui/blob/main/main.py).\nYou may also have a look at [examples.py](https://github.com/zauberzeug/nicegui/tree/main/examples.py) for more demonstrations of what you can do with NiceGUI.\n',
    'author': 'Zauberzeug GmbH',
    'author_email': 'info@zauberzeug.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/zauberzeug/nicegui',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations
from typing import Deque
import asyncio
import traceback
import urllib
from collections import deque
from justpy.htmlcomponents import WebPage
from .custom_view import CustomView
from .element import Element
from ..task_logger import create_task

class LogView(CustomView):

    def __init__(self, lines: Deque[str], max_lines: int):
        super().__init__('log', __file__, max_lines=max_lines)
        self.lines = lines
        self.allowed_events = ['onConnect']
        self.initialize(onConnect=self.handle_connect)

    def handle_connect(self, msg):
        try:
            if self.lines:
                content = '\n'.join(self.lines)
                command = f'push("{urllib.parse.quote(content)}")'
                create_task(self.run_method(command, msg.websocket), name=str(command))
        except:
            traceback.print_exc()

class Log(Element):

    def __init__(self, max_lines: int = None):
        """Log view

        Create a log view that allows to add new lines without re-transmitting the whole history to the client.

        :param max_lines: maximum number of lines before dropping oldest ones (default: None)
        """
        self.lines = deque(maxlen=max_lines)
        super().__init__(LogView(lines=self.lines, max_lines=max_lines))
        self.classes('border whitespace-pre font-mono').style('opacity: 1 !important; cursor: text !important')

    async def push_async(self, line: str):
        self.lines.append(line)
        await asyncio.gather(*[
            self.view.run_method(f'push("{urllib.parse.quote(line)}")', socket)
            for socket in WebPage.sockets.get(self.page.page_id, {}).values()
        ])

    def push(self, line: str):
        create_task(self.push_async(line), name=f'log.push line {line}')



from starlette.websockets import WebSocket
from ..task_logger import create_task


def open(self, path: str, socket: WebSocket):
    """
    Open

    Can be used to programmatically trigger redirects for a specific client.

    :param path: string that is a relative url path or an absolute url
    :param socket: WebSocket defining the target client
    """
    create_task(open_async(path, socket))

async def open_async(path: str, socket: WebSocket):
    await socket.send_json({'type': 'page_update', 'page_options': {'redirect': path}})
